package Bluestem::Client;
use strict;
use warnings;

use FileHandle ();
use DirHandle ();
use JSON::PP ();
use Digest::HMAC_SHA1 ();
use Carp qw(confess);
use Socket qw(inet_aton sockaddr_in PF_INET SOCK_STREAM);
use CGI::Cookie ();
use LWP::UserAgent ();
use HTTP::Request ();
use Frontier::RPC2 ();
my $json = JSON::PP->new();
$json->pretty(1);
$json->canonical(1);
$json->relaxed(1);

our $VERSION = '0.90.08';
our $BSVERSION = '1.6';
#our $BSRPCVERSION = '1.1';
our $BSCONF = "/var/www/bluestem-client/bluestem.conf";

sub a2h ($)
{
	my($s) = @_;
	my(@c);
	for( split(//, $s) ) { push(@c, ord($_)) };
	unpack("H*", pack("C*", @c));
}

sub new : method
{
	my($class,%p) = @_;
	my $config_file = $p{config} || $ENV{BLUESTEM_CLIENT_CONFIG} || $BSCONF;
	my $conf;
	eval {
		local $/;
		my $fh = FileHandle->new( $config_file )
			or die "Failed to open $config_file: $!\n";
		$conf = $json->decode( <$fh> );
	};
	$@ and confess "Failed to read config `$config_file': $@";
	for my $req (qw(default_domain)) {
		$conf->{$req}
			or confess "Config `$config_file' does not specify $req.";
	}

	# Check, detaint and set defaults for configuration parameters
	$conf->{cache_dir} and $conf->{cache_dir} =~ m/(.*)/
		or die "cache_dir not set in $config_file!";
	$conf->{cache_dir} = $1;
	$conf->{idle_timeout} ||= 60 * 30;
	$conf->{prior_timeout} ||= $conf->{idle_timeout};
	$conf->{servername} ||= $ENV{SERVER_NAME} || $ENV{HOSTNAME};
	$conf->{client_cookiename} ||= "Bluestem_$conf->{servername}";
	$conf->{client_cookiedomain} ||= $conf->{servername};
	$conf->{client_keysize} ||= 16;
	$conf->{logfile} ||= "/var/log/bluestem";
	$conf->{loglvl} ||= 3;
	$conf->{no_strict_ip_nets} ||= [];
	my $self = bless {
		conf => $conf,
		config_file => $config_file,
		version => $VERSION,
	}, $class;
	$self->log("Created new $class",8);
	return $self;
}

sub __get_cacheentry : method
{
	my($self,$cacheid) = @_;
	my($session_start,$cacherand) = $cacheid =~ m/^(\d+)-(\d+)$/
		or confess "Invalid cacheid format `$cacheid'";
	my $cachefile = "$self->{conf}{cache_dir}/$session_start-$cacherand";
	my $fh = FileHandle->new($cachefile);
	unless( $fh ) {
		$self->log("Failed to open cache id $cacheid: $!",5);
		return;
	}
	my @stat = stat $fh;
	my $idle = time - $stat[9]; # Idle is from mtime
	my $cacheentry = <$fh>;
	chomp($cacheentry);
	my($id,$ip,$key,$iserver) = split /,/, $cacheentry;
	my $userinfo_json = join('',<$fh>);
	my $userinfo = $userinfo_json ? $json->decode($userinfo_json) : {};
	$self->log("Retrieved cache id $cacheid $cacheid $id,$ip,****,$iserver",8);
	return($id,$ip,$key,$idle,time - $session_start,$iserver,$userinfo);
}

sub update_cacheaccess : method
{
	my($self,$cacheid) = @_;
	$cacheid =~ m/^(\d+-\d+)$/
		or die "Invalid cacheid `$cacheid'";
	$cacheid = $1;
	my $cachefile = "$self->{conf}{cache_dir}/$cacheid";
	chmod 0660, $cachefile;
	my($atime,$mtime);
	$atime = $mtime = time;
	utime $atime, $mtime, $cachefile;
	$self->log("Update cache access for cacheid $cacheid",8);
}

sub remove_cacheentry : method
{
	my($self,$cacheid) = @_;
	$cacheid =~ m/^(\d+-\d+)$/
		or die "Invalid cacheid `$cacheid'";
	$cacheid = $1;
	unlink "$self->{conf}{cache_dir}/$cacheid";
}

sub __get_session : method
{
	my($self,$cacheid,$clientkey,$clientip) = @_;
	my($id,$ip,$key,$idle,$age,$iserver,$userinfo) = $self->__get_cacheentry($cacheid)
		or return;
	
	if( $id eq 'Unknown' # Login was not completed.
	 or $clientkey ne $key
	 or !$self->ipcmp($clientip,$ip)
	 or $idle > $self->{conf}{idle_timeout}
	) {
		$self->remove_cacheentry($cacheid);
		return;
	}

	my $session = {
		netid => $id,
		cacheid => $cacheid,
		clientip => $clientip,
		idle => $idle,
		age => $age,
		iserver => $iserver,
		userinfo => $userinfo,
	};
	return $session;
}

sub bluestem_id_parse : method
{
	my($self,$id) = @_;
	return unless $id;

	my($user)   = $id =~ m{^([^@/]+)};
	my($domain) = $id =~ m{@([^/]+)};
	my($auth)   = $id =~ m{/(.*)$};

	$domain ||= $self->{conf}{default_domain};
	$auth   ||= ($domain eq $self->{conf}{default_domain})
		? $self->{conf}{default_auth}
		: 'default';

	return( $user, $domain, $auth );
}


sub bluestem_id_full : method
{
	my($self,$id) = @_;
	my($user,$domain,$auth) = $self->bluestem_id_parse($id);
	return "$user\@$domain/$auth";
}


sub bluestem_id_short : method
{
	my($self,$id) = @_;
	return unless $id;

	my $default_domain = $self->{conf}{default_domain};
	my $default_auth = $self->{conf}{default_auth};
	if ( $id =~ s{\@$default_domain(/|$)}{$1} ) {
		$id  =~ s{/$default_auth$}{};
	} else {
		$id  =~ s{/default($|@)}{$1};
	}

	return $id;
}

sub choose_id_server : method
{
	my($self,$paservers) = @_;

	# Return most recent prior auth server available if there is one.
	if( $paservers ) {
		for my $idserv ( sort {
			$paservers->{$a} <=> $paservers->{$b}
		} keys %$paservers ) {
			return $idserv
				if $self->id_server_available($idserv);
		}
	}

	# Randomly return an idserver based on weighting
	my $weightsum = 0;
	my $idshash = $self->{conf}{id_servers};
	my @id_servers = keys %$idshash;
	for my $ids ( @id_servers ) {
		$weightsum += $idshash->{$ids}{weight};
	}
	while( @id_servers ) {
		my $wn = rand $weightsum;
		my $i;
		for( $i=0; $i < @id_servers - 1; ++$i ) {
			last if 0 >= ( $wn -= $idshash->{$id_servers[$i]}{weight} );
		}
		my $ids = $id_servers[$i];
		if( $self->id_server_available($ids) ) {
			return $ids;
		}
		$weightsum -= $idshash->{$ids}{weight};
		splice @id_servers,$i,1;
	}

	die "Failed to find available id server.\n";
}

sub id_server_available : method
{
	my($self,$iserver) = @_;
	my($fqdn,$hport,$dport) = split(/:/, $iserver);

	# Just checking that https is accepting connections
	my $sfh;	
	socket( $sfh, PF_INET, SOCK_STREAM, getprotobyname("tcp") )
		or confess "Failed to open tcp socket: $!";
	my $ipaddr = inet_aton($fqdn)
		or return;
	# FIXME - add alarm or something to speed this up!
	my $portaddr = sockaddr_in(443, $ipaddr);
	connect($sfh, $portaddr)
		or return;
	return 1;
}

sub id_servers_update : method
{
	my($self,$id_servers) = @_;
	my $updated = 0;
	while( my($ids,$conf) = each %{$self->{conf}{id_servers}} ) {
		if( $id_servers->{$ids} ) {
			if( $id_servers->{$ids}{weight} != $conf->{weight} ) {
				$conf->{weight} = $id_servers->{$ids}{weight};
				$updated = 1;
			}
			if( $id_servers->{$ids}{primary} and not $conf->{primary} ) {
				$conf->{primary} = 1;
				$updated = 1;
			} elsif( $conf->{primary} and not $id_servers->{$ids}{primary} ) {
				delete $conf->{primary};
				$updated = 1;
			}
			delete $id_servers->{$ids};
		} else {
			$updated = 1;
			delete $self->{conf}{id_servers}{$ids};
		}
	}
	while ( my ($ids,$conf) = each %$id_servers ) {
		$updated = 1;
		$self->{conf}{id_servers}{$ids} = $conf;
	}

	if( $updated ) {
		$self->log("Update to id_servers: ".join(', ',keys %{$self->{conf}{id_servers}}),2);
		$self->writeconf();
	}
}

sub writeconf : method
{
	my $self = shift;
	my $fh = FileHandle->new($self->{config_file},">");
	unless( $fh ) {
		$self->log("Unable to update $self->{config_file}: $!",1);
		return;
	}
	my $text = $json->encode($self->{conf});
	$fh->print( $text,"\n" );
	$self->log("Updated $self->{config_file}",2);
	$self->log("New $self->{config_file}:\n$text",9);
}

sub generate_clientkey : method
{
	my($self,$cacheid) = @_;
	my $hmac = Digest::HMAC_SHA1->new($self->{conf}{server_key});
	$hmac->add(time,$$,rand);
	return substr($hmac->hexdigest,0,$self->{conf}{client_keysize});
}

sub create_cacheentry : method
{
	my($self,$iserver,$clientip) = @_;
	my $cacheid = sprintf("%d-%04d%d",time,int rand 9999,$$);
	while( -e "$self->{conf}{cache_dir}/$cacheid" ) {
		$cacheid = sprintf("%d-%04d%d",time,int rand 9999,$$);
	}
	my $clientkey = $self->generate_clientkey($cacheid);

	my $fh = FileHandle->new("$self->{conf}{cache_dir}/$cacheid",">");
	$fh->print("Unknown,$clientip,$clientkey,$iserver");
	$fh->close();

	chmod(0660,"$self->{conf}{cache_dir}/$cacheid");

	$self->log("Created cache entry $cacheid Unknown,$clientip,****,$iserver",5);

	return($cacheid,$clientkey);
}

sub ipcmp : method
{
	my($self,$ip1,$ip2) = @_;
	return 1 unless $self->{conf}{client_check_ip};
	return 1 if $ip1 eq $ip2;
	return 0;
}

sub update_cacheentry : method
{
	my($self,$cacheid,$netid,$newkey,$clientip,$userinfo) = @_;
	my($id,$ip,$key,$idle,$age,$iserver) = $self->__get_cacheentry($cacheid)
		or die "No such cache entry `$cacheid'\n";
	$newkey ||= $key; 
	$self->ipcmp($clientip,$ip)
		or die "IP mismatch `$clientip' != `$ip'\n";
	$cacheid =~ m/^(\d+-\d+)$/
		or die "Invalid cacheid `$cacheid'";
	$cacheid = $1;
	my $fh = FileHandle->new("$self->{conf}{cache_dir}/$cacheid",">")
		or confess "Failed to open cache entry $cacheid for writing";
	my $cache_format = $self->{conf}{cache_format} || 'new';
	if( $cache_format eq 'new' ) {
		$fh->print("$netid,$clientip,$newkey,$iserver\n");
		if( $userinfo ) {
			$fh->print($json->encode($userinfo),"\n");
		}
	} else {
		$netid = $self->bluestem_id_full($netid);
		$fh->print("$netid,$clientip,$newkey\n");
	}
	$fh->close();

	chmod(0660,"$self->{conf}{cache_dir}/$cacheid");

	$self->log("Update to cache id $cacheid to $netid,$clientip,****,$iserver",3);
}

sub login_url : method
{
	my($self,%p) = @_;
	my $return_url = $p{return_url}
		or confess "No return_url specified!";
	my $cacheid = $p{cacheid}
		or confess "No cacheid specified!";
	my $iserver = $p{iserver} || $self->choose_id_server->($p{paservers});
	my $message = $p{message} || '';
	my $prior = $p{prior} || $self->{conf}{prior_timeout} || '';

	my($fqdn,$hport,$dport) = split(/:/, $iserver);
	my $idhost = $fqdn;
	$idhost .= ":$hport" unless $hport == 443;
	my $idurl = "https://$idhost$self->{conf}{id_server_bluestem_path}/login.cgi";

	my $a2h_msg = a2h($message);
	my $a2h_retrieve = a2h("$self->{conf}{cgipath}/retrieve_auth.".($self->{conf}{is_iis}?'pl':'cgi'));
	my $a2h_return = a2h($return_url);

	my $hmac = Digest::HMAC_SHA1->new(pack("H*",$self->{conf}{server_key}));
	$hmac->add(
		"BSVersion=$Bluestem::Client::BSVERSION;".
		"cacheid=$cacheid;".
		"msg=$a2h_msg;".
		"prior=$prior;".
		"RetrieveURL=$a2h_retrieve;".
		"return=$a2h_return"
	);
	my $bsversionhash = $hmac->hexdigest();

	my $url = "$idurl?cacheid=$cacheid";
	$url .= "&return=$a2h_return";
	$url .= "&msg=$a2h_msg";
	$url .= "&prior=$prior";
	$url .= "&RetrieveURL=$a2h_retrieve";
	$url .= "&BSVersion=$Bluestem::Client::BSVERSION";
	$url .= "&BSVersionHash=$bsversionhash";

	$self->log("Returning login_url $url",9);
	return $url;
}

sub client_cookie : method
{
	my($self,$cacheid,$clientkey) = @_;
	my %cookie_param = (
		-name    => "$self->{conf}{client_cookiename}",
		-value   => "$cacheid~$clientkey",
		-secure  => 1,
	);
	$cookie_param{'-domain'} = $self->{conf}{client_cookiedomain}
		if $self->{conf}{client_cookiedomain} and $self->{conf}{client_cookiedomain} =~ m/^\./;
	$cookie_param{'-path'} = $self->{conf}{client_cookiepath}
		if $self->{conf}{client_cookiepath};
	my $cookie = CGI::Cookie->new(%cookie_param);
	my $cookiestr = $cookie->as_string;
	# Need to unencode cookie tilde for compatibility
	$cookiestr =~ s/%7E/~/g;
	$cookiestr .= '; HttpOnly';
	return $cookiestr;
}

sub idlogout_url : method
{
	my($self,%p) = @_;
	my $iserver = $p{iserver}
		or confess "No iserver specified!";
	my $return_url = $p{return_url};
	my($fqdn, $hport, $dport) = split(/:/, $iserver);
	my $host = $fqdn;
	$host .= ":$hport" unless $hport == 443;
	$return_url = a2h( $return_url );
  	my $url = "https://$host$self->{conf}{id_server_bluestem_path}/idlogout.cgi?return=$return_url";
	$self->log("Returning idlogout_url $url",9);
  	return $url;
}

sub open_log : method
{
	my($self) = @_;
	if( $self->{logfh} ) {
		$self->log("Closing log file",9);
		$self->{logfh}->close();
	}
	$self->{logfh} = FileHandle->new($self->{conf}{logfile},">>");

	unless( $self->{logfh} ) {
		warn "Failed to open bluestem log `$self->{conf}{logfile}': $!";
		return;
	}
	
	$self->{logfh}->autoflush(1);
	$self->log("Opened log file",9);
	$self->{logfh};
}

sub log : method
{
	my($self,$msg,$lvl) = @_;
	$lvl ||= 3;

	return unless $lvl <= $self->{conf}{loglvl};

	$self->{logfh} or $self->open_log()
		or return;

	my($sc,$mi,$hr,$dy,$mo,$yr,$wday) = localtime(time);
	my($pgm) = $0;
	$pgm = "[pid $$]" unless $pgm =~ m{^[/a-zA-Z0-9]};
	$pgm =~ s{^.*[/\\]}{};
	chomp $msg;
	$msg ||= "[no message]";
	
	$self->{logfh}->printf("%02d/%02d/%02d %02d:%02d:%02d %s %s: %s\n",
		$yr%100, $mo+1, $dy, $hr, $mi, $sc,
		substr('SunMonTueWedThuFriSat', 3*$wday, 3),
		$pgm, $msg,
	);
}

sub cacheclean : method
{
	my $self = shift;
	my $dh = DirHandle->new($self->{conf}{cache_dir})
		or die "Failed to open cache directory `$self->{conf}{cache_dir}': $!";
	# Reset base time so that -M will return age relative to now.
	# Used in case this is called from a long running program.
	$^T = time;
	while( my $cacheid = $dh->read ) {
		next unless $cacheid =~ m/^\d+-\d+$/;
		my $age = (-M "$self->{conf}{cache_dir}/$cacheid")
			or next;
		next unless $age * 24 * 60 * 60 > $self->{conf}{idle_timeout};
		unlink "$self->{conf}{cache_dir}/$cacheid";
	}
}

sub remove_user_cacheentry : method
{
	my($self,@netids) = @_;
	my %netids = map {($self->bluestem_id_full($_)=>1)} @netids;
	my $dh = DirHandle->new($self->{conf}{cache_dir})
		or die "Failed to open cache directory `$self->{conf}{cache_dir}': $!";
	while( my $cacheid = $dh->read ) {
		next unless $cacheid =~ m/^\d+-\d+$/;
		my($netid,$ip,$key,$idle,$age,$iserver) = $self->__get_cacheentry($cacheid);
		next unless $netids{$netid};
		$self->log("Session removed for $netid from $ip (cacheid $cacheid)",0);
		$self->remove_cacheentry($cacheid);
	}
}

sub retrieve_auth : method
{
	my($self,%p) = @_;
	my $cacheid = $p{cacheid}
		or die "No cacheid parameter provided to retrieve_auth\n";
	my $idhost = $p{idhost}
		or die "No idhost parameter provided to retrieve_auth\n";
	my $apphost = $p{apphost}
		or die "No apphost parameter provided to retrieve_auth\n";
	my $credentialkey = $p{credentialkey}
		or die "No credentialkey parameter provided to retrieve_auth\n";

	my  $request = HTTP::Request->new();
	my  $rpc     = Frontier::RPC2->new();

	$request->header('Content-Type' => 'text/xml');
	$request->method('POST');
	$request->uri("https://$idhost$self->{conf}{id_server_bluestem_path}/verify_auth.cgi"); 
	$request->user_agent('retrieve_auth 1.6');

	my  $vfyhash = Digest::HMAC_SHA1::hmac_sha1_hex("$cacheid;$apphost;$credentialkey",$self->{conf}{server_key});

	$request->content($rpc->encode_call('Bluestem.VerifyAuth',$cacheid,$apphost,$credentialkey,$vfyhash));

	my $ua = LWP::UserAgent->new();

	my $response = $ua->request($request);
	if ($response->is_error) {
		die "Error calling Bluestem.VerifyAuth: ".$response->status_line."\n";
	}

	# Handle Frontier's mis-handling of <i8> by just dropping the data type
	my $content = $response->content();
	$content =~ s{</?i8>}{}g;

	my  $rpcresp = $rpc->decode($content);
	if ($rpcresp->{type} eq 'fault') {
    		my $faultstring = $rpcresp->{value}[0]{faultString};
		die "Error calling Bluestem.VerifyAuth: ".$faultstring."\n";
	}

	my $auth = $rpcresp->{value}[0]
		or die "No authentication data in ID Server response!\n";

	$credentialkey eq Digest::HMAC_SHA1::hmac_sha1_hex("$cacheid;$auth->{loginToken}",$self->{conf}{server_key})
		or die "Invalid loginToken in RPC response!\n";

	$self->update_cacheentry($cacheid,$auth->{bluestemID},$auth->{loginToken},$p{clientip},$auth->{userInfo});

	return $auth;
}

1;
